/* Copyright (C) 2013-2021 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

#define _POSIX_C_SOURCE 200112L

#include "endianness.h"
#include "hash.h"
#include "math.h"
#include "mem_allocator.h"

#include <string.h>

struct buffer {
  const char* mem;
  size_t len;
};

/* Array of round constants: first 32 bits of the fractional parts of the cube
 * roots of the first 64 primes 2..311 */
static const uint32_t k[64] = {
  0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1,
  0x923f82a4, 0xab1c5ed5, 0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
  0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174, 0xe49b69c1, 0xefbe4786,
  0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
  0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147,
  0x06ca6351, 0x14292967, 0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
  0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85, 0xa2bfe8a1, 0xa81a664b,
  0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
  0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a,
  0x5b9cca4f, 0x682e6ff3, 0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
  0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
};

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
/* Right rotation */
static FINLINE uint32_t
rrot(const uint32_t ui, const unsigned int count)
{
  ASSERT(count <= 32);
  return ui >> count | ui << (32 - count);
}

static void
get_chunk512(char dst[64], const size_t ichunk512, void* ctx)
{
  struct buffer* buf = ctx;
  const char* chunk = NULL;
  size_t chunk_sz = 0;
  const size_t offset = ichunk512*(sizeof(char[64]));
  ASSERT(buf && offset < buf->len);

  chunk = buf->mem + offset;
  chunk_sz = MMIN(buf->len - offset, sizeof(char[64]));
  memcpy(dst, chunk, chunk_sz);
}

/*******************************************************************************
 * Exported functions
 ******************************************************************************/
void
hash_sha256
  (const void* data,
   const size_t len,
   hash256_T hash)
{
  struct chunked_data_desc desc = CHUNKED_DATA_DESC_NULL;
  struct buffer buf;

  buf.mem = data;
  buf.len = len;

  desc.get_chunk512 = get_chunk512;
  desc.size = len;
  desc.context = &buf;

  hash_sha256_chunked_data(&desc, hash);
}

void
hash_sha256_chunked_data
  (struct chunked_data_desc* data,
   hash256_T hash)
{
  /* Actually, the size of one chunk is 64 bytes. Anyway, note that after the
   * last byte of the data to digest, one has to store 9 additionnal bytes: 1
   * to mark the end of the data and 8 used to save the effective data length.
   * Such tailing bytes can exceed the 64 bytes length of the chunk.  Since the
   * overall message length must be aligned on 64 bytes we thus pre-allocate an
   * array whose size is 2 times the chunk size */
  char buffer[64*2];

  size_t msg_sz = 0;
  size_t ichunk = 0;
  size_t nchunks = 0;
  size_t ichunk_last_byte = 0; /* Chunk id the last data byte */
  size_t remaining_data_sz = 0;

  /* Initial hash values: first 32 bits of the fractional parts of the square
   * roots of the first 8 primes 2..19 */
  uint32_t h0 = 0x6a09e667;
  uint32_t h1 = 0xbb67ae85;
  uint32_t h2 = 0x3c6ef372;
  uint32_t h3 = 0xa54ff53a;
  uint32_t h4 = 0x510e527f;
  uint32_t h5 = 0x9b05688c;
  uint32_t h6 = 0x1f83d9ab;
  uint32_t h7 = 0x5be0cd19;
  ASSERT(data && hash);

  remaining_data_sz = data->size;

  /* Compute the overall size of the message */
  msg_sz = ALIGN_SIZE
    (data->size + 1/*Byte of the '1' bit*/ + 8/*message len*/, 64u);

  ASSERT((msg_sz % 64) == 0);
  nchunks = msg_sz / 64; /* #chunks of 512 bits */

  /* Index of the chunk containing the last data byte */
  ichunk_last_byte = data->size / 64;

  FOR_EACH(ichunk, 0, nchunks) {
    char* chunk = NULL;
    uint32_t w[64] = {0};
    uint32_t a=h0, b=h1, c=h2, d=h3, e=h4, f=h5, g=h6, h=h7;
    int i;

    if(ichunk > ichunk_last_byte) {
      chunk = buffer + 64;
    } else {
      chunk = buffer;

      /* Fetch the user data if any */
      if(remaining_data_sz) {
        ASSERT(data->get_chunk512);
        data->get_chunk512(buffer, ichunk, data->context);
        remaining_data_sz = remaining_data_sz > 64
          ? remaining_data_sz - 64 : 0;
      }

      /* Setup the tailing bytes of the message */
      if(ichunk == ichunk_last_byte) {
        char* remaining_msg = buffer;
        const size_t remaining_msg_sz = ichunk_last_byte==nchunks-1 ? 64 : 128;

        /* Id after the last data byte */
        const size_t ibyte_len = data->size % 64;

        /* Setup the '1' bit that marks the end of the data */
        remaining_msg[ibyte_len] = (char)0x80u;

        /* Clean up the bytes after the data up to the last 8 bytes */
        memset(remaining_msg + ibyte_len + 1, 0,
          remaining_msg_sz - ibyte_len - 1 - 8);

        /* Store the message 8*len in big endian */
        *(uint64_t*)(remaining_msg + remaining_msg_sz-8) =
          big_endian_64(data->size*8);
      }
    }

    FOR_EACH(i, 0, 16) {
      w[i] = big_endian_32(((uint32_t*)chunk)[i]);
    }
    FOR_EACH(i, 16, 64) {
      const uint32_t s0 = rrot(w[i-15],7) ^ rrot(w[i-15],18) ^ (w[i-15] >> 3);
      const uint32_t s1 = rrot(w[i-2],17) ^ rrot(w[i-2], 19) ^ (w[i-2]  >> 10);
      w[i] = w[i-16] + s0 + w[i-7] + s1;
    }

    /* Compress the chunk */
    FOR_EACH(i, 0, 64) {
      const uint32_t s1 = rrot(e, 6) ^ rrot(e, 11) ^ rrot(e, 25);
      const uint32_t ch = (e & f) ^ ((~e) & g);
      const uint32_t tmp1 = h + s1 + ch + k[i] + w[i];
      const uint32_t s0 = rrot(a, 2) ^ rrot(a, 13) ^ rrot(a, 22);
      const uint32_t maj = (a & b) ^ (a & c) ^ (b & c);
      const uint32_t tmp2 = s0 + maj;

      h = g;
      g = f;
      f = e;
      e = d + tmp1;
      d = c;
      c = b;
      b = a;
      a = tmp1 + tmp2;
    }

    /* Update hash with the current compressed chunk */
    h0+=a; h1+=b; h2+=c; h3+=d; h4+=e; h5+=f; h6+=g; h7+=h;
  }

  /* Write the results */
  ((uint32_t*)hash)[0] = big_endian_32(h0);
  ((uint32_t*)hash)[1] = big_endian_32(h1);
  ((uint32_t*)hash)[2] = big_endian_32(h2);
  ((uint32_t*)hash)[3] = big_endian_32(h3);
  ((uint32_t*)hash)[4] = big_endian_32(h4);
  ((uint32_t*)hash)[5] = big_endian_32(h5);
  ((uint32_t*)hash)[6] = big_endian_32(h6);
  ((uint32_t*)hash)[7] = big_endian_32(h7);
}

void
hash256_to_cstr(const hash256_T hash, char cstr[65])
{
  size_t i;
  ASSERT(hash && cstr);
  FOR_EACH(i, 0, sizeof(hash256_T)) {
    sprintf(cstr+i*2, "%02x", (uint8_t)hash[i]);
  }
}

int
hash256_eq(const hash256_T hash0, const hash256_T hash1)
{
  return memcmp(hash0, hash1, sizeof(hash256_T)) == 0;
}

