/* Copyright (C) 2015-2021 |Meso|Star> (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#ifndef TEST_SSP_RAN_SPHERICAL_ZONE_H
#define TEST_SSP_RAN_SPHERICAL_ZONE_H

#include "ssp.h"
#include "test_ssp_utils.h"

#define NSAMPS 1024

#endif /* TEST_SSP_RAN_SPHERE_H */

#if TYPE_FLOAT==0
  #define REAL double
  #define TEST test_double
  #define RAN_SPHERICAL_ZONE_UNIFORM_LOCAL ssp_ran_spherical_zone_uniform_local
  #define RAN_SPHERICAL_ZONE_UNIFORM_PDF ssp_ran_spherical_zone_uniform_pdf
  #define RAN_SPHERICAL_ZONE_UNIFORM ssp_ran_spherical_zone_uniform
  #define RAN_SPHERE_UNIFORM ssp_ran_sphere_uniform
  #define EQ_EPS eq_eps
  #define R3_EQ_EPS d3_eq_eps
  #define R3_IS_NORMALIZED d3_is_normalized
  #define R3_NORMALIZE d3_normalize
  #define R3_DOT d3_dot
#elif TYPE_FLOAT==1
  #define REAL float
  #define TEST test_float
  #define RAN_SPHERICAL_ZONE_UNIFORM_LOCAL ssp_ran_spherical_zone_uniform_float_local
  #define RAN_SPHERICAL_ZONE_UNIFORM_PDF ssp_ran_spherical_zone_uniform_float_pdf
  #define RAN_SPHERICAL_ZONE_UNIFORM ssp_ran_spherical_zone_uniform_float
  #define RAN_SPHERE_UNIFORM ssp_ran_sphere_uniform_float
  #define EQ_EPS eq_epsf
  #define R3_EQ_EPS f3_eq_eps
  #define R3_IS_NORMALIZED f3_is_normalized
  #define R3_NORMALIZE f3_normalize
  #define R3_DOT f3_dot
#else
  #error "TYPE_FLOAT must be defined either 0 or 1"
#endif

static void
TEST(void)
{
  struct ssp_rng* rng;
  struct mem_allocator allocator;
  REAL pdf;
  REAL samps[NSAMPS][3];
  REAL up[3];
  REAL heights[2];
  int i, j;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);
  CHK(ssp_rng_create(&allocator, &ssp_rng_mt19937_64, &rng) == RES_OK);

  heights[0] = heights[1] = 1;
  CHK(RAN_SPHERICAL_ZONE_UNIFORM_LOCAL(rng, heights, samps[0], &pdf) == samps[0]);
  CHK(samps[0][0] == 0);
  CHK(samps[0][1] == 0);
  CHK(samps[0][2] == 1);
  CHK(IS_INF(pdf));

  heights[0] = -1;
  CHK(RAN_SPHERICAL_ZONE_UNIFORM_LOCAL(rng, heights, samps[0], &pdf) == samps[0]);
  CHK(EQ_EPS(pdf, 1/(4*(REAL)PI), (REAL)1.e-6));

  heights[0] = heights[1] = (REAL)0.4;
  CHK(RAN_SPHERICAL_ZONE_UNIFORM_LOCAL(rng, heights, samps[0], &pdf) == samps[0]);
  CHK(EQ_EPS(pdf, 1/(2*(REAL)PI), (REAL)1.e-6));

  /* Check NULL pdf */
  CHK(RAN_SPHERICAL_ZONE_UNIFORM_LOCAL(rng, heights, samps[0], NULL) == samps[0]);

  FOR_EACH(i, 0, NSAMPS) {
    heights[0] = (REAL)-0.7; heights[1] = 1;
    CHK(RAN_SPHERICAL_ZONE_UNIFORM_LOCAL(rng, heights, samps[i], &pdf) == samps[i]);
    CHK(R3_IS_NORMALIZED(samps[i]));
    CHK(EQ_EPS(pdf, 1/(2*(REAL)PI*(heights[1]-heights[0])), (REAL)1.e-6));
    CHK(EQ_EPS(pdf, RAN_SPHERICAL_ZONE_UNIFORM_PDF(heights), (REAL)1.e-6));
    CHK(samps[i][2] >= heights[0] && samps[i][2] <= heights[1]);
    FOR_EACH(j, 0, i) {
      CHK(!R3_EQ_EPS(samps[j], samps[i], (REAL)1.e-6));
    }
  }

  /* Sample an up vector */
  RAN_SPHERE_UNIFORM(rng, up, NULL);

  heights[0] = heights[1] = 1;
  CHK(RAN_SPHERICAL_ZONE_UNIFORM(rng, up, heights, samps[0], &pdf) == samps[0]);
  CHK(R3_EQ_EPS(samps[0], up, (REAL)1.e-6));
  CHK(IS_INF(pdf));

  heights[0] = -1;
  CHK(RAN_SPHERICAL_ZONE_UNIFORM(rng, up, heights, samps[0], &pdf) == samps[0]);
  CHK(EQ_EPS(pdf, 1/(4*(REAL)PI), (REAL)1.e-6));

  heights[0] = heights[1] = (REAL)0.9;
  CHK(RAN_SPHERICAL_ZONE_UNIFORM(rng, up, heights, samps[0], &pdf) == samps[0]);
  CHK(EQ_EPS(pdf, 1/(2*(REAL)PI), (REAL)1.e-6));

  /* Check NULL pdf */
  CHK(RAN_SPHERICAL_ZONE_UNIFORM(rng, up, heights, samps[0], NULL) == samps[0]);

  FOR_EACH(i, 0, NSAMPS) {
    heights[0] = (REAL)0.3; heights[1] = 1;
    CHK(RAN_SPHERICAL_ZONE_UNIFORM(rng, up, heights, samps[i], &pdf) == samps[i]);
    CHK(R3_IS_NORMALIZED(samps[i]));
    CHK(EQ_EPS(pdf, 1/(2*(REAL)PI*(heights[1]-heights[0])), (REAL)1.e-6));
    CHK(EQ_EPS(pdf, RAN_SPHERICAL_ZONE_UNIFORM_PDF(heights), (REAL)1.e-6));
    CHK(R3_DOT(up, samps[i]) >= heights[0] && R3_DOT(up, samps[i]) <= heights[1]);
    FOR_EACH(j, 0, i) {
      CHK(!R3_EQ_EPS(samps[j], samps[i], (REAL)1.e-6));
    }
  }

  ssp_rng_ref_put(rng);
  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);

  CHK(mem_allocated_size() == 0);
}

#undef REAL
#undef TEST
#undef RAN_SPHERICAL_ZONE_UNIFORM_LOCAL
#undef RAN_SPHERICAL_ZONE_UNIFORM_PDF
#undef RAN_SPHERICAL_ZONE_UNIFORM
#undef RAN_SPHERE_UNIFORM
#undef EQ_EPS
#undef R3_EQ_EPS
#undef R3_IS_NORMALIZED
#undef R3_NORMALIZE
#undef R3_DOT
#undef TYPE_FLOAT
